/*
 * Copyright (c) 2012, Texas Instruments Incorporated
 * http://www.ti.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
/*
 *  ======== OpenMP.xs ========
 */

var OpenMP        = null;
var MultiProc     = null;
var SharedRegion  = null;
var Ipc           = null;

/*
 *  ======== module$meta$init ========
 */
function module$meta$init()
{
    /* Only process during "cfg" phase */
    if (xdc.om.$name != "cfg") {
        return;
    }

    OpenMP = this;
}

/*
 *  ======== module$use ========
 */
function module$use()
{
    // Setup IPC between the cores participating in the runtime. This is
    // required to initialize a SharedRegion which contains a shared heap,
    // HeapMemMP. When the OpenMP runtime is in RTSC mode, the HeapMemMP is
    // used to support malloc's from shared memory.
    configureIpc(OpenMP.masterCoreIdx, OpenMP.numCores);
   
    // If OpenMP thread stacks are placed in L2SRAM, the runtime will set
    // SP to the global address E.g. on Core 0, 0x800000 is converted to
    // 0x10800000
    // This conversion breaks the initialization routine for BIOS stack check - 
    // ti_sysbios_hal_Hwi_initStack. E.g., on Core 0, STACK_BASE is 0x800000, 
    // local stack is 0x10800000 and MSMC is overwritten by:
    // while (curStack-- > (UArg)STACK_BASE) 
    if (OpenMP.allocateStackFromHeap == false)
    {
        var Hwi = xdc.useModule('ti.sysbios.hal.Hwi');
        Hwi.initStackFlag = false;
        var Task = xdc.useModule('ti.sysbios.knl.Task');
        Task.checkStackFlag = false;
        Task.initStackFlag = false;
    }

    // Reserve HW Semaphores used by the OpenMP runtime
    var GateHWSem       = xdc.useModule('ti.sdo.ipc.gates.GateHWSem');
    var ompSettings = xdc.module('ti.runtime.openmp.Settings');
    if (ompSettings.usingOpenCL == true)
    {
        GateHWSem.setReserved(0);
        GateHWSem.setReserved(1);
        GateHWSem.setReserved(2);
    }
    for (var i = 0; i < OpenMP.hwSemCount; i++) 
    {
        GateHWSem.setReserved(OpenMP.hwSemBaseIdx + i);
    }

    //Capture CPU clock frequency information from the platform file. 
    //clockFreq is used by the OpenMP runtime timing functions: 
    //omp_get_wtime() and omp_get_wtick()
    OpenMP.clockFreq = Program.cpu.clockRate;

}


/*
 *  ======== module$static$init ========
 */
function module$static$init(mod, params)
{
    mod.useSharedHeap = false;
}

/*
 *  ======== configureIpc ========
 */
function configureIpc(masterCoreIdx, numCores)
{
    var IpcSettings = xdc.useModule('ti.sdo.ipc.family.Settings');
    var procNames   = IpcSettings.getDeviceProcNames().sort();

    // On heterogenous SoCs, exclude HOST cores. The DSPs are named
    // CORE0 through COREn
    var dspNames = [];
    for (coreIdx in procNames) {
       if (procNames[coreIdx].match("CORE"))
           dspNames.push(procNames[coreIdx]);
    }


    if (dspNames.length < numCores) {
        throw("ERROR: Too many processors requested: " + numCores);
    }

    // Configure MultiProc with the base index and number of cores
    var MultiProc   = xdc.useModule('ti.sdo.utils.MultiProc');
    MultiProc.baseIdOfCluster = masterCoreIdx;
    // Total number of processors
    MultiProc.numProcessors = numCores;
    // Number of processors in cluster set by setConfig
    MultiProc.setConfig(null,
                 dspNames.slice(MultiProc.baseIdOfCluster, numCores));

    // Avoid wasting shared memory for MessageQ transports or notify.
    // Ipc is used exclusively to set up a HeapMemMP in the specified
    // Shared Region to support a shared heap for malloc's
    var Ipc         = xdc.useModule('ti.sdo.ipc.Ipc');
    for (var i = 0; i < MultiProc.numProcessors; i++) {
        Ipc.setEntryMeta({
                remoteProcId: i,
                setupMessageQ: false,
                setupNotify: false,
        });
    }

    // Attach and synchronize all cores configured via MultiProc
    Ipc.procSync = Ipc.ProcSync_ALL;
}


/*
 *  ======== module$validate ========
 */
function module$validate()
{
    var ompSettings = xdc.module('ti.runtime.openmp.Settings');
    if (ompSettings.usingOpenCL == false)
    {
        if (OpenMP.numCores == 0)
            OpenMP.$logFatal("Set numCores in config file", OpenMP);

        if (OpenMP.clockFreq == 0)
            OpenMP.$logFatal("Set CPU clock rate in Platform file", OpenMP);

        if (OpenMP.msmcBase == 0)
            OpenMP.$logFatal("Set msmcBase in config file", OpenMP);

        if (OpenMP.msmcNoCachePhysicalBase == 0)
            OpenMP.$logFatal("Set msmcNoCacheBase in config file", OpenMP);

        if (OpenMP.msmcSize == 0)
            OpenMP.$logFatal("Set msmcSize in config file", OpenMP);

        if (OpenMP.msmcNoCacheVirtualBase == 0)
            OpenMP.$logFatal("Set msmcNoCacheBase in config file", OpenMP);

        if (OpenMP.msmcNoCacheVirtualSize == 0)
            OpenMP.$logFatal("Set msmcNoCacheSize in config file", OpenMP);

        if (OpenMP.ddrBase == 0)
            OpenMP.$logFatal("Set ddrBase in config file", OpenMP);

        if (OpenMP.ddrSize == 0)
            OpenMP.$logFatal("Set ddrSize in config file", OpenMP);

        if ((OpenMP.allocateStackFromHeap == true) && 
           (OpenMP.allocateStackFromHeapSize == 0))
            OpenMP.$logFatal("Stack size cannot be 0", OpenMP);
    }
}
